@@ -7,7 +7,8 @@
 #define __COMPAT_H__
 
 #ifndef HAVE_TPARM
-char *tparm (const char *, ...);
+char *my_tparm (const char *, ...);
+#define tparm my_tparm
 #endif
 
 #ifndef HAVE_STRTOUL
@@ -22,9 +23,13 @@ size_t	strlcpy (char *, const char *, size_t);
 size_t	strlcat (char *, const char *, size_t);
 #endif
 
-void	bsd_arc4random_stir (void);
-void	bsd_arc4random_addrandom (unsigned char *, int);
+#ifndef HAVE_ARC4RANDOM
 u_32int_t	bsd_arc4random (void);
+#define arc4random bsd_arc4random
+#endif
+
+int	my_base64_encode (const void *, int, char **);
+int	my_base64_decode (const char *, void *);
 
 #ifndef HAVE_VSNPRINTF
 int	vsnprintf (char *, size_t, const char *, va_list);
@@ -50,4 +55,25 @@ int	unsetenv (const char *);
 char *	my_realpath (const char *, char x[MAXPATHLEN]);
 #endif
 
+/*
+#ifndef HAVE_STRTOIMAX
+#ifdef strtoimax
+#undef strtoimax
+#endif
+long	strtoimax (const char *, char **, int);
+#endif
+*/
+
+#if 0
+/* humanize_number(3) */ 
+#define HN_DECIMAL              0x01
+#define HN_NOSPACE              0x02
+#define HN_B                    0x04
+#define HN_DIVISOR_1000         0x08
+
+#define HN_GETSCALE             0x10
+#define HN_AUTOSCALE            0x20
+int	humanize_number (char *, size_t, long, const char *, int, int);
+#endif
+
 #endif
