@@ -1,4 +1,4 @@
-/* $EPIC: keys.c,v 1.32 2005/04/20 04:07:02 wd Exp $ */
+/* $EPIC: keys.c,v 1.59 2010/01/08 05:15:03 jnelson Exp $ */
 /*
  * keys.c:  Keeps track of what happens whe you press a key.
  *
@@ -35,7 +35,7 @@
 #include "config.h"
 #include "commands.h"
 #include "functions.h"
-#include "history.h"
+#include "hook.h"
 #include "ircaux.h"
 #include "input.h"
 #include "keys.h"
@@ -44,10 +44,46 @@
 #include "output.h"
 #include "screen.h"
 #include "stack.h"
-#include "term.h"
+#include "termx.h"
 #include "vars.h"
 #include "window.h"
 #include "timer.h"
+#include "reg.h"
+#include "server.h"
+
+/* This is a typedef for a function used with the /BIND system.  The
+ * functions all live in input.c.  Following it is a macro to quickly define
+ * functions to handle keybindings. */
+typedef void (*BindFunction) (unsigned int, unsigned char *);
+
+/* This is the structure used to hold binding functions. */
+struct Binding {
+    struct Binding *next;   /* linked-list stuff. :) */
+
+    char    *name;          /* the name of this binding */
+    BindFunction func;      /* function to use ... */
+    char    *alias;         /* OR alias to call.  one or the other. */
+    char    *filename;      /* the package which added this binding */
+};
+extern struct Binding *binding_list; /* list head for keybindings */
+
+struct Binding  *add_binding    (const char *, BindFunction, char *);
+void            remove_binding  (char *);
+struct Binding *find_binding    (const char *);
+
+#define KEYMAP_SIZE 256
+struct Key {
+    unsigned char val;  /* the key value */
+    unsigned char changed; /* set if this binding was changed post-startup */
+    struct Binding *bound; /* the function we're bound to. */
+    struct Key *map;    /* a map of subkeys (may be NULL) */
+    struct Key *owner;  /* the key which contains the map we're in. */
+    char    *stuff;     /* 'stuff' associated with our binding */
+    char    *filename;  /* the package which added this binding */
+};
+
+extern struct Key *head_keymap; /* the head keymap.  the root of the keys.  */
+static int bind_compressed_string (unsigned char *cs, int slen, const char *bindstr, char *args);
 
 /* This file is split into two pieces.  The first piece represents bindings.
  * Bindings are now held in a linked list, allowing the user to add new ones
@@ -94,32 +130,33 @@ struct Binding *add_binding (const char *name, BindFunction func, char *alias) {
 	bp->func = func;
 	bp->alias = NULL;
     }
-    bp->filename = malloc_strdup(current_package());
+    bp->filename = NULL;
+    if (current_package())
+        bp->filename = malloc_strdup(current_package());
 
     add_to_list((List **)&binding_list, (List *)bp);
 
     return bp;
 }
 
-
 static void remove_bound_keys(struct Key *, struct Binding *);
 void remove_binding (char *name) {
     struct Binding *bp;
 
     if (!name)
-        return;
+	return;
 
     bp = (struct Binding *)remove_from_list((List **)&binding_list, name);
     if (bp) {
         /* be sure to remove any keys bound to this binding first */
         remove_bound_keys(head_keymap, bp);
 
-        new_free(&bp->name);
-        if (bp->alias)
-            new_free(&bp->alias);
-        if (bp->filename)
-            new_free(&bp->filename);
-        new_free(&bp);
+	new_free(&bp->name);
+	if (bp->alias)
+	    new_free(&bp->alias);
+	if (bp->filename)
+	    new_free(&bp->filename);
+	new_free(&bp);
     }
 
     return;
@@ -129,14 +166,13 @@ static void remove_bound_keys (struct Key *map, struct Binding *binding) {
     int c;
 
     for (c = 0; c <= KEYMAP_SIZE - 1;c++) {
-        if (map[c].bound == binding)
+	if (map[c].bound == binding)
             map[c].bound = NULL;
-        if (map[c].map)
+	if (map[c].map)
             remove_bound_keys(map[c].map, binding);
     }
 }
 
-
 struct Binding *find_binding (const char *name) {
     if (!name)
 	return NULL;
@@ -150,13 +186,11 @@ void init_binds (void) {
     ADDBIND("ALTCHARSET",		    insert_altcharset		    );
     ADDBIND("BACKSPACE",		    input_backspace		    );
     ADDBIND("BACKWARD_CHARACTER",	    backward_character		    );
-    ADDBIND("BACKWARD_HISTORY",		    backward_history		    );
     ADDBIND("BACKWARD_WORD",		    input_backward_word		    );
     ADDBIND("BEGINNING_OF_LINE",	    input_beginning_of_line	    );
     ADDBIND("BLINK",			    insert_blink		    );
     ADDBIND("BOLD",			    insert_bold			    );
-    ADDBIND("CLEAR_SCREEN",		    clear_screen		    );
-    ADDBIND("COMMAND_COMPLETION",	    command_completion		    );
+    ADDBIND("CLEAR_SCREEN",		    my_clear_screen		    );
     ADDBIND("CPU_SAVER",		    cpu_saver_on		    );
     ADDBIND("DELETE_CHARACTER",		    input_delete_character	    );
     ADDBIND("DELETE_NEXT_WORD",		    input_delete_next_word	    );
@@ -167,7 +201,6 @@ void init_binds (void) {
     ADDBIND("ERASE_TO_BEG_OF_LINE",	    input_clear_to_bol		    );
     ADDBIND("ERASE_TO_END_OF_LINE",	    input_clear_to_eol		    );
     ADDBIND("FORWARD_CHARACTER",	    forward_character		    );
-    ADDBIND("FORWARD_HISTORY",		    forward_history		    );
     ADDBIND("FORWARD_WORD",		    input_forward_word		    );
     ADDBIND("HIGHLIGHT_OFF",		    highlight_off		    );
     ADDBIND("NEXT_WINDOW",		    next_window			    );
@@ -176,8 +209,9 @@ void init_binds (void) {
     ADDBIND("QUIT_IRC",			    irc_quit			    );
     ADDBIND("QUOTE_CHARACTER",		    quote_char			    );
     ADDBIND("REFRESH_INPUTLINE",	    refresh_inputline		    );
-    ADDBIND("REFRESH_SCREEN",		    (BindFunction) refresh_screen   );
+    ADDBIND("REFRESH_SCREEN",		    refresh_screen   		    );
     ADDBIND("REFRESH_STATUS",		    (BindFunction) update_all_status);
+    ADDBIND("RESET_LINE",		    input_reset_line		    );
     ADDBIND("REVERSE",			    insert_reverse		    );
     ADDBIND("SCROLL_BACKWARD",		    scrollback_backwards	    );
     ADDBIND("SCROLL_END",		    scrollback_end		    );
@@ -185,12 +219,12 @@ void init_binds (void) {
     ADDBIND("SCROLL_START",		    scrollback_start		    );
     ADDBIND("SELF_INSERT",		    input_add_character		    );
     ADDBIND("SEND_LINE",		    send_line			    );
-    ADDBIND("SHOVE_TO_HISTORY",		    shove_to_history		    );
     ADDBIND("STOP_IRC",			    term_pause			    );
     ADDBIND("SWAP_LAST_WINDOW",		    swap_last_window		    );
     ADDBIND("SWAP_NEXT_WINDOW",		    swap_next_window		    );
     ADDBIND("SWAP_PREVIOUS_WINDOW",	    swap_previous_window	    );
     ADDBIND("SWITCH_CHANNELS",		    switch_channels		    );
+    ADDBIND("SWITCH_QUERY",		    switch_query		    );
     ADDBIND("TOGGLE_INSERT_MODE",	    toggle_insert_mode		    );
     ADDBIND("TOGGLE_STOP_SCREEN",	    toggle_stop_screen		    );
     ADDBIND("TRANSPOSE_CHARACTERS",	    input_transpose_characters	    );
@@ -212,7 +246,7 @@ struct Key	*construct_keymap	(struct Key *);
 int		clean_keymap		(struct Key *);
 unsigned char	*bind_string_compress	(const unsigned char *, int *);
 unsigned char	*bind_string_decompress	(unsigned char *, const unsigned char *, int);
-static int	bind_string		(const u_char *, const char *, char *);
+static int	bind_string		(const unsigned char *, const char *, char *);
 struct Key	*find_sequence		(const unsigned char *, int);
 void		show_all_bindings	(struct Key *, const unsigned char *, size_t);
 void		show_all_rbindings	(struct Key *, const unsigned char *, int, struct Binding *);
@@ -247,17 +281,19 @@ static void key_exec (struct Key *key) {
 	return;
     }
 
-    /* check alias first, then function */
-    if (key->bound->alias != NULL) {
-	/* I don't know if this is right ... */
-	char *exec = malloc_strdup(key->bound->alias);
-	if (key->stuff)
-	    malloc_strcat_wordlist(&exec, " ", key->stuff);
-	parse_line(NULL, exec, empty_string, 0, 0);
-	new_free(&exec);
-    } else if (key->bound->func != NULL)
-	key->bound->func(key->val, key->stuff);
-
+    if (do_hook(KEYBINDING_LIST, "%s 0 %d", key->bound->name, (int)key->val))
+    {
+	    /* check alias first, then function */
+	    if (key->bound->alias != NULL) {
+		/* I don't know if this is right ... */
+		char *exec = malloc_strdup(key->bound->alias);
+		if (key->stuff)
+		    malloc_strcat_wordlist(&exec, " ", key->stuff);
+		call_lambda_command("KEY", exec, empty_string);
+		new_free(&exec);
+	    } else if (key->bound->func != NULL)
+		key->bound->func(key->val, key->stuff);
+    }
     return;
 }
 	
@@ -265,16 +301,19 @@ static void key_exec (struct Key *key) {
  * into a string, and then parses the string looking for the longest
  * possible input combinations and executing them as it goes. */
 void key_exec_bt (struct Key *key) {
-    unsigned char *kstr = empty_string, *nstr;
+    const unsigned char *kstr = (const unsigned char *)empty_string;
+    const unsigned char *nstr;
     int len = 1, kslen;
     struct Key *kp;
 
     /* now walk backwards, growing kstr as necessary */
     while (key != NULL) {
-	nstr = alloca(len + 1);
-	memcpy(nstr + 1, kstr, len);
-	nstr[0] = key->val;
-	kstr = nstr;
+        unsigned char *buf;
+
+	buf = alloca(len + 1);
+	memcpy(buf + 1, kstr, len);
+	buf[0] = key->val;
+	kstr = buf;
 	len++;
 	key = key->owner;
     }
@@ -322,11 +361,11 @@ void key_exec_bt (struct Key *key) {
  * previous key's action (if there has been a timeout).  The timeout factor
  * is set in milliseconds by the KEY_INTERVAL variable.  See further for
  * instructions. :) */
-struct Key *handle_keypress (struct Key *last, Timeval pressed, unsigned char key) {
-    struct Key *kp;
-    
+void *	handle_keypress (void *lastp, Timeval pressed, unsigned char key) {
+    struct Key *kp, *last;
+
     /* we call the timeout code here, too, just to be safe. */
-    last = timeout_keypress(last, pressed);
+    last = timeout_keypress(lastp, pressed);
 
     /* if last is NULL (meaning we're in a fresh state), pull from the head
      * keymap.  if last has a map, pull from that map.  if last has no map,
@@ -344,28 +383,30 @@ struct Key *handle_keypress (struct Key *last, Timeval pressed, unsigned char ke
     /* If there is a map and a keybinding, schedule a timeout */
     if (kp->map && kp->bound)
 	add_timer(0, empty_string, get_int_var(KEY_INTERVAL_VAR) / 1000.0, 1,
-			do_input_timeouts, NULL, NULL, -1);
+			do_input_timeouts, NULL, NULL, GENERAL_TIMER, -1, 0);
 
     /* if the key has a map associated, we can't automatically execute the
      * action.  return kp and wait quietly. */
     if (kp->map != NULL)
-	return kp;
+	return (void *)kp;
 
     /* otherwise, we can just exec our key and return nothing. */
     key_exec(kp);
     return NULL;
 }
 
-struct Key *timeout_keypress (struct Key *last, Timeval pressed) {
+void *	timeout_keypress (void *lastp, Timeval pressed) {
     int mpress = 0; /* ms count since last pressing */
     Timeval tv;
     Timeval right_now;
+    struct Key *last;
 
-    if (last == NULL)
+    if (lastp == NULL)
 	return NULL; /* fresh state, we need not worry about timeouts */
 
+    last = (struct Key *)lastp;
     if (last->bound == NULL)
-	return last; /* wait unconditionally if this key is unbound. */
+	return (void *)last; /* wait unconditionally if this key is unbound. */
 
     get_time(&right_now);
     tv = time_subtract(pressed, right_now);
@@ -378,9 +419,43 @@ struct Key *timeout_keypress (struct Key *last, Timeval pressed) {
 	key_exec(last);
 	return NULL; /* we're no longer waiting on this key's map */
     }
-    return last; /* still waiting.. */
+    return (void *)last; /* still waiting.. */
 }
 
+/* do_input_timeouts: walk through the list of screens and potentially do
+ * timeouts for input */
+ /* XXX This doesn't belong here -- maybe in keys.c? */
+int     do_input_timeouts (void *ignored)
+{
+    Screen *oldscreen = last_input_screen;
+    Screen *screen;
+    int server;
+
+    if (!foreground)
+        return 0;               /* Not a lot of input.. :) */
+
+    /* I hope this is sufficient to allow input to be processed.  I took
+     * a lot of the guts from do_screens().  Here's hoping I didn't
+     * break anything. ;) */
+    for (screen = screen_list; screen; screen = screen->next) {
+        if (screen->alive) {
+            server = from_server;
+            last_input_screen = screen;
+            output_screen = screen;
+            make_window_current(screen->current_window);
+            from_server = current_window->server;
+            screen->last_key = timeout_keypress(screen->last_key,
+					        screen->last_press);
+            from_server = server;
+        }
+    }
+    output_screen = last_input_screen = oldscreen;
+    return 0;
+}
+
+
+/**************************************************************************/
+
 struct Key *construct_keymap (struct Key *owner) {
     int c;
     struct Key *map = new_malloc(sizeof(struct Key) * KEYMAP_SIZE);
@@ -404,7 +479,7 @@ struct Key *construct_keymap (struct Key *owner) {
  * are not left around in the timeout system.  the function returns positive
  * if *the map passed* was removed, negative otherwise. */
 int clean_keymap (struct Key *map) {
-    int c;
+    int	c;
     int save = 0;
 
     /* walk through the map to see if things are in use.  if something is
@@ -526,8 +601,8 @@ unsigned char *bind_string_compress (const unsigned char *str, int *len) {
 
 /* this decompresses a compressed bind string into human-readable form.  it
  * assumes sufficient memory has already been allocated for it. */
-unsigned char *bind_string_decompress (unsigned char *dst, const unsigned char
-	*src, int srclen) {
+unsigned char *	bind_string_decompress (unsigned char *dst, const unsigned char *src, int srclen) 
+{
     unsigned char *ret = dst;
 
     while (srclen) {
@@ -547,20 +622,34 @@ unsigned char *bind_string_decompress (unsigned char *dst, const unsigned char
     return ret;
 }
 
+static	int	bind_string (const unsigned char *sequence, const char *bindstr, char *args) 
+{
+	unsigned char *cs;
+	int	slen, retval;
+
+	cs = bind_string_compress(sequence, &slen);
+	if (cs == NULL) {
+		yell("bind_string(): couldn't compress sequence %s", sequence);
+		return 0;
+	}
+
+	retval = bind_compressed_string(cs, slen, bindstr, args);
+	new_free(&cs);
+	return retval;
+}
+
 /* this function takes a key sequence (user-input style), a function to bind
  * to, and optionally arguments to that function, and does all the work
  * necessary to bind it.  it will create new keymaps as it goes, if
  * necessary, etc. */
-static int bind_string (const unsigned char *sequence, const char *bindstr, char *args) {
-    unsigned char *cs; /* the compressed keysequence */
+static int bind_compressed_string (unsigned char *cs, int slen, const char *bindstr, char *args) {
     unsigned char *s;
-    int slen;
     struct Key *kp = NULL;
     struct Key *map = head_keymap;
     struct Binding *bp = NULL;
 
-    if (!sequence || !bindstr) {
-	yell("bind_string(): called without sequence or bind function!");
+    if (!cs || !slen || !bindstr) {
+	yell("bind_compressed_string(): called without sequence or bind function!");
 	return 0;
     }
 
@@ -571,12 +660,6 @@ static int bind_string (const unsigned char *sequence, const char *bindstr, char
 	return 0;
     }
 
-    cs = bind_string_compress(sequence, &slen);
-    if (cs == NULL) {
-	yell("bind_string(): couldn't compress sequence %s", sequence);
-	return 0;
-    }
-
     s = cs;
     while (slen) {
 	kp = &map[*s++];
@@ -599,7 +682,8 @@ static int bind_string (const unsigned char *sequence, const char *bindstr, char
 	    if (bp != NULL) {
 		if (args)
 		    kp->stuff = malloc_strdup(args);
-		kp->filename = malloc_strdup(current_package());
+	        if (current_package())
+		    malloc_strcpy(&kp->filename, current_package());
 	    }
 	}
     }
@@ -607,7 +691,6 @@ static int bind_string (const unsigned char *sequence, const char *bindstr, char
     /* if we're post-initialization, clean out the keymap with each call. */
     if (bind_post_init)
 	clean_keymap(head_keymap);
-    new_free(&cs);
     return 1;
 }
 
@@ -658,7 +741,7 @@ void init_keys (void) {
     /* bind characters 32 - 255 to SELF_INSERT. */
     s[1] = '\0';
     for (c = 32;c <= KEYMAP_SIZE - 1;c++) {
-	s[0] = c;
+	s[0] = (unsigned char )c;
 	BIND(s, "SELF_INSERT");
     }
 
@@ -676,9 +759,7 @@ void init_keys (void) {
     BIND("^K", "ERASE_TO_END_OF_LINE");
     BIND("^L", "REFRESH_SCREEN");
     BIND("^M", "SEND_LINE");
-    BIND("^N", "FORWARD_HISTORY");
     BIND("^O", "HIGHLIGHT_OFF");
-    BIND("^P", "BACKWARD_HISTORY");
     BIND("^Q", "QUOTE_CHARACTER");
     /* ^R */
     BIND("^S", "TOGGLE_STOP_SCREEN");
@@ -691,7 +772,6 @@ void init_keys (void) {
     BIND("^Z", "STOP_IRC");
     /* ^[ (was META1_CHARACTER) */
     /* ^\ */
-    BIND("^]", "SHOVE_TO_HISTORY");
     /* ^^ */
     BIND("^_", "UNDERLINE");
     /* mind the gap .. */
@@ -709,7 +789,6 @@ void init_keys (void) {
 #endif
 
     /* now for what was formerly meta1 (escape) sequences. */
-    BIND("^[^[", "COMMAND_COMPLETION");
     BIND("^[.", "CLEAR_SCREEN");
     BIND("^[<", "SCROLL_START");
     BIND("^[>", "SCROLL_END");
@@ -726,10 +805,6 @@ void init_keys (void) {
     /* meta2 stuff. */
     BIND("^[O^Z", "STOP_IRC");
     BIND("^[[^Z", "STOP_IRC");
-    BIND("^[OA", "BACKWARD_HISTORY");
-    BIND("^[[A", "BACKWARD_HISTORY");
-    BIND("^[OB", "FORWARD_HISTORY");
-    BIND("^[[B", "FORWARD_HISTORY");
     BIND("^[OC", "FORWARD_CHARACTER");
     BIND("^[[C", "FORWARD_CHARACTER");
     BIND("^[OD", "BACKWARD_CHARACTER");
@@ -771,10 +846,6 @@ void init_termkeys (void) {
 }
 
     bind_post_init = 0;
-    TBIND(key_up, BACKWARD_HISTORY);
-    TBIND(key_down, FORWARD_HISTORY);
-    TBIND(key_left, BACKWARD_CHARACTER);
-    TBIND(key_right, FORWARD_CHARACTER);
     TBIND(key_ppage, SCROLL_BACKWARD);
     TBIND(key_npage, SCROLL_FORWARD);
     TBIND(key_home, SCROLL_START);
@@ -785,66 +856,32 @@ void init_termkeys (void) {
 #undef TBIND
 }
 
-/* save_bindings is called by the /save command to..well.. save bindings.
- * we call the save_bindings_recurse() function which acts a lot like
- * (surprise surprise) show_all_bindings/show_key in tandem. */
-void save_bindings_recurse (FILE *, struct Key *, const unsigned char *, size_t);
-void save_bindings (FILE *fp, int do_all) {
-    save_bindings_recurse(fp, head_keymap, "", 0);
-}
-
-void save_bindings_recurse (FILE *fp, struct Key *map, const unsigned char *str, size_t len) {
-    int c;
-    unsigned char *newstr;
-    unsigned char *ds; /* decompressed sequence */
-    size_t size;
-
-    size = len + 2;
-    newstr = alloca(size);
-    strlcpy(newstr, str, size);
-    ds = alloca(((len + 1) * 2) + 1);
-
-    /* go through our map, see what is changed, and save it.  recurse down
-     * as necessary. */
-    newstr[len + 1] = '\0';
-    for (c = 0; c <= KEYMAP_SIZE - 1;c++) {
-	newstr[len] = c;
-	if (map[c].bound && map[c].changed) {
-	    bind_string_decompress(ds, newstr, len + 1);
-	    fprintf(fp, "BIND %s %s%s%s\n", ds, map[c].bound->name,
-		    (map[c].stuff ? " " : ""),
-		    (map[c].stuff ? map[c].stuff : ""));
-	}
-	if (map[c].map)
-	    save_bindings_recurse(fp, map[c].map, newstr, len + 1);
-    }
-}
-
 /* this is called only by irc_exit, and its purpose is to free
  * all our allocated stuff. */
-void remove_bindings_recurse (struct Key *);
+void remove_bindings_recurse (struct Key **);
 void remove_bindings (void) {
 
     while (binding_list != NULL)
 	remove_binding(binding_list->name);
 
-    remove_bindings_recurse(head_keymap);
+    remove_bindings_recurse(&head_keymap);
 }
 
-void remove_bindings_recurse (struct Key *map) {
+void remove_bindings_recurse (struct Key **mapptr) {
+    struct Key *map = *mapptr;
     int c;
 
     /* go through our map, clear any memory that might be left lying around.
      * recurse as necessary */
     for (c = 0; c <= KEYMAP_SIZE - 1;c++) {
 	if (map[c].map)
-	    remove_bindings_recurse(map[c].map);
+	    remove_bindings_recurse(&(map[c].map));
 	if (map[c].stuff)
 	    new_free(&map[c].stuff);
 	if (map[c].filename)
 	    new_free(&map[c].filename);
     }
-    new_free(&map);
+    new_free((char **)mapptr);
 }
 
 /* this is called when a package is unloaded.  we should unset any
@@ -858,7 +895,7 @@ void unload_bindings (const char *pkg) {
     bp = binding_list;
     while (bp != NULL) {
 	bp2 = bp->next;
-	if (!my_stricmp(bp->filename, pkg))
+	if (bp->filename && !my_stricmp(bp->filename, pkg))
 	    remove_binding(bp->name);
 	bp = bp2;
     }
@@ -874,7 +911,7 @@ void unload_bindings_recurse (const char *pkg, struct Key *map) {
     for (c = 0; c <= KEYMAP_SIZE - 1;c++) {
 	/* if the key is explicitly bound to something, and it was done in
 	 * our package, unbind it. */
-	if (map[c].bound && !my_stricmp(map[c].filename, pkg)) {
+	if (map[c].bound && map[c].filename && !my_stricmp(map[c].filename, pkg)) {
 	    if (map[c].stuff)
 		new_free(&map[c].stuff);
 	    if (map[c].filename)
@@ -889,15 +926,19 @@ void unload_bindings_recurse (const char *pkg, struct Key *map) {
 /* set_key_interval:  this is used to construct a new Timeval when the
  * 'KEY_INTERVAL' /set is changed.  We modify an external variable which
  * defines how long the client will wait to timeout, at most. */
-void set_key_interval (const void *stuff) {
-    int msec = *(const int *)stuff;
+void set_key_interval (void *stuff) {
+    VARIABLE *v;
+    int msec;
+
+    v = (VARIABLE *)stuff;
+    msec = v->integer;
 
     if (msec < 10) {
 	say("Setting KEY_INTERVAL below 10ms is not recommended.");
 	msec = 10;
     }
 
-    set_int_var(KEY_INTERVAL_VAR, msec);
+    v->integer = msec;
 }
 
 /* do_stack_bind:  this handles the /stack .. bind command.  below is the
@@ -1042,6 +1083,8 @@ BUILT_IN_COMMAND(bindcmd) {
     const unsigned char *seq;
     char *function;
     int recurse = 0;
+    unsigned char *cs;
+    int	slen, retval;
 
     if ((seq = new_next_arg(args, &args)) == NULL) {
 	show_all_bindings(head_keymap, "", 0);
@@ -1051,6 +1094,8 @@ BUILT_IN_COMMAND(bindcmd) {
     /* look for flags */
     if (*seq == '-') {
 	if (!my_strnicmp(seq + 1, "DEFAULTS", 1)) {
+	    remove_bindings();
+	    init_binds();
 	    init_keys();
 	    init_termkeys();
 	    return;
@@ -1074,28 +1119,32 @@ BUILT_IN_COMMAND(bindcmd) {
     }
 
     if ((function = new_next_arg(args, &args)) == NULL) {
-	unsigned char *compstr;
-	int slen;
-	compstr = bind_string_compress(seq, &slen);
-	if (compstr == NULL)
+	cs = bind_string_compress(seq, &slen);
+	if (cs == NULL)
 	    return; /* umm.. */
 
-	show_key(NULL, compstr, slen, recurse);
+	show_key(NULL, cs, slen, recurse);
+	return;
+    }
+
+    if (!(cs = bind_string_compress(seq, &slen))) {
+	yell("BIND: couldn't compress sequence %s", seq);
 	return;
     }
 
     /* bind_string() will check any errors for us. */
-    if (!bind_string(seq, function, *args ? args : NULL)) {
+    retval = bind_compressed_string(cs, slen, function, *args ? args : NULL);
+    if (!retval) {
 	if (!my_strnicmp(function, "meta", 4))
 	    yell(
 "Please note that the META binding functions are no longer available.  \
 For more information please see the bind(4) helpfile and the file \
 doc/keys distributed with the EPIC source."
 		);
-		    
-	return; /* assume an error was spouted for us. */
     }
-    show_key(NULL, seq, 0, 0);
+    else
+        show_key(NULL, cs, slen, 0);
+    new_free(&cs);
 }
 
 /* support function for /bind:  this function shows, recursively, all the
@@ -1138,7 +1187,7 @@ void show_key (struct Key *key, const unsigned char *str, int slen, int recurse)
     else {
 	if (bp) {
 	    say("[%s] \"%s\" is bound to %s%s%s",
-		    (*key->filename ? key->filename : "*"),
+		    (empty(key->filename) ? "*" : key->filename),
 		    (slen ? bind_string_decompress(clean, str, slen) : str), 
 		    bp->name,
 		    (key->stuff ? " " : ""), (key->stuff ? key->stuff : ""));
@@ -1217,7 +1266,7 @@ BUILT_IN_COMMAND(parsekeycmd) {
 	    fake.stuff = malloc_strdup(args);
 	else
 	    fake.stuff = NULL;
-	fake.filename = empty_string;
+	fake.filename = LOCAL_COPY("");
 
 	key_exec(&fake);
 
@@ -1255,19 +1304,23 @@ char *bindctl (char *input)
     char *listc;
     char *retval = NULL;
 
-    GET_STR_ARG(listc, input);
+    GET_FUNC_ARG(listc, input);
     if (!my_strnicmp(listc, "FUNCTION", 1)) {
 	struct Binding *bp;
 	char *func;
 
-	GET_STR_ARG(func, input);
+	GET_FUNC_ARG(func, input);
 	bp = find_binding(func);
 
-	GET_STR_ARG(listc, input);
+	GET_FUNC_ARG(listc, input);
 	if (!my_strnicmp(listc, "CREATE", 1)) {
 	    char *alias;
 
-	    GET_STR_ARG(alias, input);
+	    /* XXX Ugh, for backwards compatability */
+	    if (*input == '"')
+		GET_DWORD_ARG(alias, input)
+	    else
+		alias = input;
 
 	    if (bp) {
 		if (bp->func)
@@ -1304,14 +1357,14 @@ char *bindctl (char *input)
 	    len = strlen(func);
 	    for (bp = binding_list;bp;bp = bp->next) {
 		if (!my_strnicmp(bp->name, func, len))
-		    malloc_strcat_word(&retval, space, bp->name);
+		    malloc_strcat_word(&retval, space, bp->name, DWORD_NO);
 	    }
 
 	    RETURN_STR(retval);
 	} else if (!my_strnicmp(listc, "PMATCH", 1)) {
 	    for (bp = binding_list;bp;bp = bp->next) {
 		if (wild_match(func, bp->name))
-		    malloc_strcat_word(&retval, space, bp->name);
+		    malloc_strcat_word(&retval, space, bp->name, DWORD_NO);
 	    }
 
 	    RETURN_STR(retval);
@@ -1330,9 +1383,9 @@ char *bindctl (char *input)
 	struct Key *key;
 	unsigned char *seq;
 
-	GET_STR_ARG(seq, input);
+	GET_DWORD_ARG(seq, input);
 	key = find_sequence(seq, 0);
-	GET_STR_ARG(listc, input);
+	GET_FUNC_ARG(listc, input);
 	if (!my_stricmp(listc, "GET")) {
 	    if (key == NULL || key->bound == NULL)
 		RETURN_EMPTY;
@@ -1342,7 +1395,7 @@ char *bindctl (char *input)
 		malloc_strcat_wordlist(&retval, " ", key->stuff);
 	    RETURN_STR(retval);
 	} else if (!my_stricmp(listc, "SET")) {
-	    GET_STR_ARG(listc, input);
+	    GET_FUNC_ARG(listc, input);
 
 	    RETURN_INT(bind_string(seq, listc, (*input ? input : NULL)));
 	} else if (!my_strnicmp(listc, "GETPACKAGE", 4)) {
@@ -1382,8 +1435,7 @@ char *bindctl (char *input)
 	} else if (!my_strnicmp(listc, "CLEAR", 1)) {
 	    if (key == NULL || key->map == NULL)
 		RETURN_INT(0);
-	    remove_bindings_recurse(key->map);
-	    key->map = NULL;
+	    remove_bindings_recurse(&key->map);
 	    RETURN_INT(1);
 	}
     }
